import logging
import xml.etree.ElementTree as ET

import gpxpy.gpx

from .data_filters import (
    discard_resample_gps_data,
    linear_resample_gps_data,
    lpf_resample_gps_data,
)
from .metadata_manager import extract_gps_info
from .mp4_manager import MP4Manager

logger = logging.getLogger(__name__)  # pylint: disable=C0103


class OsmoGps:
    gps_data = None
    inputs = None
    input_frame_rate = None
    output_frequency = None
    resampling_method = None
    extract_extensions = False

    def __init__(self, inputs, timezone_offset=0, extract_extensions=False):
        if inputs is None:
            raise ValueError("inputs cannot be None")
        self.inputs = inputs
        self.timezone_offset = timezone_offset
        self.extract_extensions = extract_extensions

        self.extract()

    def extract(self):

        logger.info(f"Running extract command with inputs: {self.inputs}")

        self.gps_data = []
        for i, input_file in enumerate(self.inputs, start=1):
            logger.info(f"Processing file {i}/{len(self.inputs)}: {input_file}")

            mp4 = MP4Manager(input_file)
            metadata = mp4.get_metadata()

            gps_info, input_frame_rate = extract_gps_info(
                metadata, self.timezone_offset, self.extract_extensions
            )
            logger.info(f"Frame rate: {input_frame_rate}")
            self.input_frame_rate = input_frame_rate
            logger.info(f"Extracted {len(gps_info)} GPS data points.")

            self.gps_data.extend(gps_info)

    def resample(
        self,
        output_frequency=None,
        resampling_method=None,
    ):
        self.resampling_method = resampling_method
        if self.resampling_method is not None:
            if self.resampling_method not in ["discard", "linear", "lpf", "none"]:
                raise ValueError(
                    "resampling_method must be one of "
                    "'discard', 'linear', 'lpf', 'none'"
                )
            self.output_frequency = output_frequency
            if self.resampling_method in ["discard", "linear", "lpf"]:
                if self.output_frequency is None:
                    raise ValueError(
                        "output_frequency cannot be None when "
                        "resampling_method is not 'none'"
                    )
        logger.info(
            f"Resampling GPS data with method: {self.resampling_method}, "
            f"output frequency: {self.output_frequency}"
        )
        if self.resampling_method is not None:
            if self.resampling_method == "linear":
                resampled_data = linear_resample_gps_data(
                    self.gps_data, self.input_frame_rate, self.output_frequency
                )
            elif self.resampling_method == "lpf":
                resampled_data = lpf_resample_gps_data(
                    self.gps_data, self.input_frame_rate, self.output_frequency
                )
            elif self.resampling_method == "discard":
                resampled_data = discard_resample_gps_data(
                    self.gps_data, self.input_frame_rate, self.output_frequency
                )
            self.gps_data = resampled_data

    def save_gpx(self, output_file):
        if self.gps_data != []:
            gpx = gpxpy.gpx.GPX()
            gpx.creator = "pyosmogps -- https://github.com/francescocaponio/pyosmogps"
            track = gpxpy.gpx.GPXTrack()
            gpx.tracks.append(track)
            segment = gpxpy.gpx.GPXTrackSegment()
            track.segments.append(segment)

            latitude = [point["latitude"] for point in self.gps_data]
            longitude = [point["longitude"] for point in self.gps_data]
            altitude = [point["altitude"] for point in self.gps_data]
            timeinfo = [point["timeinfo"] for point in self.gps_data]
            if self.extract_extensions:
                camera_acc_x = [point["camera_acc_x"] for point in self.gps_data]
                camera_acc_y = [point["camera_acc_y"] for point in self.gps_data]
                camera_acc_z = [point["camera_acc_z"] for point in self.gps_data]
                remote_der_x = [point["remote_der_x"] for point in self.gps_data]
                remote_der_y = [point["remote_der_y"] for point in self.gps_data]
                remote_der_z = [point["remote_der_z"] for point in self.gps_data]

            for i in range(len(timeinfo)):
                point = gpxpy.gpx.GPXTrackPoint(
                    latitude=latitude[i],
                    longitude=longitude[i],
                    elevation=altitude[i],
                    time=timeinfo[i],
                )

                if self.extract_extensions:
                    extensions = ET.Element("extensions")

                    acc_x_ext = ET.SubElement(extensions, "acc_x")
                    acc_x_ext.text = f"{camera_acc_x[i]:.3f}"

                    acc_y_ext = ET.SubElement(extensions, "acc_y")
                    acc_y_ext.text = f"{camera_acc_y[i]:.3f}"

                    acc_z_ext = ET.SubElement(extensions, "acc_z")
                    acc_z_ext.text = f"{camera_acc_z[i]:.3f}"

                    der_x_ext = ET.SubElement(extensions, "der_x")
                    der_x_ext.text = f"{remote_der_x[i]:.3f}"

                    der_y_ext = ET.SubElement(extensions, "der_y")
                    der_y_ext.text = f"{remote_der_y[i]:.3f}"

                    der_z_ext = ET.SubElement(extensions, "der_z")
                    der_z_ext.text = f"{remote_der_z[i]:.3f}"

                    point.extensions.append(extensions)
                segment.points.append(point)

            with open(output_file, "w") as gpx_file:
                gpx_file.write(gpx.to_xml())

            logger.info(f"GPS data written to {output_file}")
            return True
        else:
            logger.info("No GPS data extracted.")
            return False

    def get_altitude(self):
        return [point["altitude"] for point in self.gps_data]

    def get_latitude(self):
        return [point["latitude"] for point in self.gps_data]

    def get_longitude(self):
        return [point["longitude"] for point in self.gps_data]



from datetime import timedelta

import numpy as np
from scipy.signal import butter, filtfilt


def discard_resample_gps_data(gps_info, input_frequency, output_frequency):
    """
    Resample the GPS data by discarding samples.

    :param gps_info: List of dicts containing GPS data.
    :param input_frequency: Original frequency of the GPS data (Hz).
    :param output_frequency: Desired frequency of the GPS data (Hz).
    :return: Resampled list of dicts.
    """
    if output_frequency > input_frequency:
        raise ValueError("Output frequency cannot be higher than input frequency.")

    # Calculate the sampling step
    step = int(input_frequency / output_frequency)
    if step < 1:
        raise ValueError("Invalid step size. Check input and output frequencies.")

    # Subsample the data
    resampled_data = gps_info[::step]
    return resampled_data


def lpf_resample_gps_data(gps_info, input_frequency, output_frequency):
    """
    Resample the GPS data using a low pass filter method.

    :param gps_info: List of dicts containing GPS data.
    :param input_frequency: Original frame rate of the GPS data (Hz).
    :param output_frequency: Desired frequency of the GPS data (Hz).
    :return: Resampled list of dicts.
    """

    # Calculate cutoff frequency
    cutoff_frequency = output_frequency / 4.0

    # Design the Butterworth low-pass filter
    nyquist = 0.5 * input_frequency
    normal_cutoff = cutoff_frequency / nyquist
    b, a = butter(4, normal_cutoff, btype="low", analog=False)

    # Apply the filter to each field
    resampled_data = []

    original_timestamps = [entry["timeinfo"] for entry in gps_info]
    total_duration = (original_timestamps[-1] - original_timestamps[0]).total_seconds()
    num_samples = int(total_duration * output_frequency)
    new_timestamps = [
        original_timestamps[0] + timedelta(seconds=i / output_frequency)
        for i in range(num_samples)
    ]

    original_timestamps_seconds = np.array(
        [(t - original_timestamps[0]).total_seconds() for t in original_timestamps]
    )

    for new_time in new_timestamps:
        new_entry = {"timeinfo": new_time}

        for key in gps_info[0].keys():
            if key == "timeinfo":
                continue

            # Extract values for the current key
            values = np.array([entry[key] for entry in gps_info])

            # Apply the low-pass filter
            filtered_values = filtfilt(b, a, values)

            # Interpolate the filtered data to match the new timestamps
            new_value = np.interp(
                (new_time - original_timestamps[0]).total_seconds(),
                original_timestamps_seconds,
                filtered_values,
            )

            new_entry[key] = new_value

        resampled_data.append(new_entry)

    return resampled_data


def linear_resample_gps_data(gps_info, input_frequency, output_frequency):
    """
    Resample the GPS data using a linear interpolation method.

    :param gps_info: List of dicts containing GPS data.
    :param input_frequency: Original frame rate of the GPS data (Hz).
    :param output_frequency: Desired frequency of the GPS data (Hz).
    :return: Resampled list of dicts.
    """

    # Calculate total duration and the new sample interval
    total_duration = (
        gps_info[-1]["timeinfo"] - gps_info[0]["timeinfo"]
    ).total_seconds()
    num_samples = int(total_duration * output_frequency)
    new_timestamps = [
        gps_info[0]["timeinfo"] + timedelta(seconds=i / output_frequency)
        for i in range(num_samples)
    ]

    resampled_data = []

    # Extract original timestamps and values
    original_timestamps = [entry["timeinfo"] for entry in gps_info]
    original_timestamps_seconds = np.array(
        [(t - original_timestamps[0]).total_seconds() for t in original_timestamps]
    )

    # Interpolate each field
    for new_time in new_timestamps:
        new_entry = {"timeinfo": new_time}

        for key in gps_info[0].keys():
            if key == "timeinfo":
                continue

            # Extract values for the current key
            values = np.array([entry[key] for entry in gps_info])

            # Interpolate linearly
            new_value = np.interp(
                (new_time - original_timestamps[0]).total_seconds(),
                original_timestamps_seconds,
                values,
            )

            new_entry[key] = new_value

        resampled_data.append(new_entry)

    return resampled_data



from typing import NamedTuple

from .pyosmogps import OsmoGps  # noqa: F401

__package_name__ = "pyosmogps"


class _VersionInfo(NamedTuple):
    major: int
    minor: int
    micro: int
    releaselevel: str
    serial: int

    @property
    def version_str(self) -> str:
        release_level = f".{self.releaselevel}" if self.releaselevel else ""
        return f"{self.major}.{self.minor}.{self.micro}{release_level}"


# Leave version as a placeholder,
# it will be updated by check_package_version.py
# during the build process.
__version_info__ = _VersionInfo(
    major=0,
    minor=2,
    micro=0,
    releaselevel='gh-action',
    serial=0,
)

__version__ = __version_info__.version_str
VERSION = __version__  # synonym for backwards compatibility



import argparse
import logging
import logging.config
import sys

from . import OsmoGps
from . import __version__ as pyosmogps_version

logger = logging.getLogger(__name__)  # pylint: disable=C0103


def _make_parser() -> argparse.ArgumentParser:
    # Separated so sphinx-argparse-cli can do its auto documentation magic.
    parser = argparse.ArgumentParser(
        description="Extract the GPS data from the Osmo Action video files",
        prog="pyosmogps",
    )
    parser.add_argument(
        "command",
        choices=["extract", "merge"],
        help="Specify the command to run: 'extract' to extract "
        "GPS data or 'merge' to merge GPX files.",
    )
    parser.add_argument(
        "inputs",
        nargs="+",
        help="Input file(s). Accepts a single file or multiple files.",
    )
    parser.add_argument(
        "output",
        help="Output file. Accepts a single file or multiple files.",
    )
    parser.add_argument(
        "--frequency",
        "-f",
        type=float,
        default=2.0,
        help="Set the output data frequency in Hz (default: 2 Hz).",
    )
    parser.add_argument(
        "--resampling-method",
        "-r",
        choices=["discard", "linear", "lpf", "none"],
        default="linear",
        help="Set the method for resampling data: 'discard' to drop "
        "excess samples, 'linear' for linear interpolation, 'lpf' "
        "for low pass filtering, 'none' for no data reduction (default: linear).",
    )
    parser.add_argument(
        "--timezone-offset",
        "-t",
        type=int,
        default=0,
        help="Set the timezone offset in hours (default: 0).",
    )
    parser.add_argument(
        "--version", "-v", action="version", version=f"%(prog)s {pyosmogps_version}"
    )
    return parser


def extract(inputs, output, frequency, resampling_method, timezone_offset=0) -> bool:
    try:
        gps = OsmoGps(inputs, timezone_offset)
        gps.resample(frequency, resampling_method)
        gps.save_gpx(output)

    except Exception as e:
        logger.error(f"Error: {e}")
        return False
    return True


def main() -> int:
    parser = _make_parser()
    if len(sys.argv) < 2:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    if args.command == "extract":
        if not args.inputs or not args.output:
            parser.error(
                "'extract' command requires at least one input file and one "
                "output file."
            )
        success = extract(
            args.inputs,
            args.output,
            args.frequency,
            args.resampling_method,
            args.timezone_offset,
        )
        return 0 if success else 1

    elif args.command == "merge":
        print("Running merge command...")
        # TODO: Implement merge command
    else:
        parser.print_help()
        parser.exit()

    return 0


if __name__ == "__main__":
    exit(main())



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: dji.proto
# Protobuf Python Version: 5.29.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC, 5, 29, 3, "", "dji.proto"
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\tdji.proto"\x87\x01\n\rDjiModuleInfo\x12\x12\n\nproto_name\x18\x01 \x01(\t\x12\x0c\n\x04unk1\x18\x02 \x01(\t\x12\x0c\n\x04unk2\x18\x03 \x01(\t\x12\x15\n\rserial_number\x18\x05 \x01(\t\x12\x0c\n\x04unk4\x18\x06 \x01(\t\x12\x0c\n\x04unk5\x18\t \x01(\r\x12\x13\n\x0b\x63\x61mera_name\x18\n \x01(\t"9\n\x12\x44jiVideoGlobalInfo\x12#\n\x0bmodule_info\x18\x01 \x03(\x0b\x32\x0e.DjiModuleInfo"W\n\x16\x44jiRemoteGpsDeviceInfo\x12\x0c\n\x04unk1\x18\x01 \x01(\r\x12\x0c\n\x04unk2\x18\x02 \x01(\r\x12\x13\n\x0b\x64\x65vice_name\x18\x04 \x01(\t\x12\x0c\n\x04unk3\x18\x05 \x01(\x02"\\\n\x1b\x44jiRemoteGpsCoordinatesInfo\x12\x18\n\x10\x63oordinate_units\x18\x01 \x01(\r\x12\x10\n\x08latitude\x18\x02 \x01(\x01\x12\x11\n\tlongitude\x18\x03 \x01(\x01"\'\n\x13\x43oordinatesDatetime\x12\x10\n\x08\x64\x61tetime\x18\x01 \x01(\t"\xa2\x01\n\x17\x44jiRemoteGpsCoordinates\x12*\n\x04info\x18\x01 \x01(\x0b\x32\x1c.DjiRemoteGpsCoordinatesInfo\x12\x17\n\x0fgps_altitude_mm\x18\x02 \x01(\r\x12\x0c\n\x04unk2\x18\x04 \x01(\r\x12\x0c\n\x04unk3\x18\x05 \x01(\r\x12&\n\x08\x64\x61tetime\x18\x06 \x01(\x0b\x32\x14.CoordinatesDatetime"6\n\x13\x44jiRemoteDerivative\x12\t\n\x01x\x18\x01 \x01(\x02\x12\t\n\x01y\x18\x02 \x01(\x02\x12\t\n\x01z\x18\x03 \x01(\x02"\x9a\x01\n\x10\x44jiRemoteGpsInfo\x12,\n\x0b\x64\x65vice_info\x18\x01 \x01(\x0b\x32\x17.DjiRemoteGpsDeviceInfo\x12-\n\x0b\x63oordinates\x18\x02 \x01(\x0b\x32\x18.DjiRemoteGpsCoordinates\x12)\n\x0b\x64\x65rivatives\x18\x03 \x01(\x0b\x32\x14.DjiRemoteDerivative" \n\rDjiUnkfloat_1\x12\x0f\n\x07unknown\x18\x01 \x01(\x02" \n\rDjiUnkInt32_1\x12\x0f\n\x07unknown\x18\x01 \x01(\x05"!\n\x0e\x44jiUnkUInt32_1\x12\x0f\n\x07unknown\x18\x01 \x01(\r")\n\x10\x44jiUnkRepBytes_1\x12\x15\n\runknown_field\x18\x01 \x03(\x0c"B\n\x12\x44jiCameraAccelInfo\x12\x0b\n\x03unk\x18\x01 \x01(\x02\x12\t\n\x01x\x18\x02 \x01(\x02\x12\t\n\x01y\x18\x03 \x01(\x02\x12\t\n\x01z\x18\x04 \x01(\x02"\xcc\x02\n\rDjiCameraInfo\x12\x1c\n\x04unk1\x18\x02 \x01(\x0b\x32\x0e.DjiUnkfloat_1\x12\x1c\n\x04unk2\x18\x03 \x01(\x0b\x32\x0e.DjiUnkfloat_1\x12\x1f\n\x04unk3\x18\x04 \x01(\x0b\x32\x11.DjiUnkRepBytes_1\x12\x1c\n\x04unk4\x18\x05 \x01(\x0b\x32\x0e.DjiUnkfloat_1\x12\x1d\n\x04unk5\x18\x06 \x01(\x0b\x32\x0f.DjiUnkUInt32_1\x12\x1d\n\x04unk6\x18\x07 \x01(\x0b\x32\x0f.DjiUnkUInt32_1\x12\x0c\n\x04unk7\x18\x08 \x01(\r\x12+\n\x0e\x61\x63\x63\x65lerometer1\x18\t \x01(\x0b\x32\x13.DjiCameraAccelInfo\x12+\n\x0e\x61\x63\x63\x65lerometer2\x18\n \x01(\x0b\x32\x13.DjiCameraAccelInfo\x12\x0c\n\x04unk8\x18\x0b \x01(\r\x12\x0c\n\x04unk9\x18\x0c \x01(\r"1\n\x0cGpsFrameInfo\x12\x10\n\x08\x66rame_id\x18\x01 \x01(\r\x12\x0f\n\x07unknown\x18\x02 \x01(\r"\x80\x01\n\nDjiGpsInfo\x12!\n\nframe_info\x18\x01 \x01(\x0b\x32\r.GpsFrameInfo\x12#\n\x0b\x63\x61mera_info\x18\x02 \x01(\x0b\x32\x0e.DjiCameraInfo\x12*\n\x0fremote_gps_info\x18\x04 \x01(\x0b\x32\x11.DjiRemoteGpsInfo"\x83\x01\n\x16VideoStreamInfoDetails\x12\r\n\x05width\x18\x01 \x01(\r\x12\x0e\n\x06height\x18\x02 \x01(\r\x12\x12\n\nframe_rate\x18\x03 \x01(\x02\x12\x0c\n\x04unk1\x18\x04 \x01(\r\x12\x0c\n\x04unk2\x18\x05 \x01(\r\x12\x0c\n\x04unk3\x18\x06 \x01(\r\x12\x0c\n\x04unk5\x18\x08 \x01(\r"I\n\x0fVideoStreamInfo\x12\x0c\n\x04type\x18\x01 \x01(\t\x12(\n\x07\x64\x65tails\x18\x03 \x01(\x0b\x32\x17.VideoStreamInfoDetails"\x8c\x01\n\x0eGenericMessage\x12.\n\x11video_global_info\x18\x01 \x01(\x0b\x32\x13.DjiVideoGlobalInfo\x12+\n\x11video_stream_info\x18\x02 \x01(\x0b\x32\x10.VideoStreamInfo\x12\x1d\n\x08gps_info\x18\x03 \x03(\x0b\x32\x0b.DjiGpsInfob\x06proto3'  # noqa: E501
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "dji_pb2", _globals)
if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals["_DJIMODULEINFO"]._serialized_start = 14
    _globals["_DJIMODULEINFO"]._serialized_end = 149
    _globals["_DJIVIDEOGLOBALINFO"]._serialized_start = 151
    _globals["_DJIVIDEOGLOBALINFO"]._serialized_end = 208
    _globals["_DJIREMOTEGPSDEVICEINFO"]._serialized_start = 210
    _globals["_DJIREMOTEGPSDEVICEINFO"]._serialized_end = 297
    _globals["_DJIREMOTEGPSCOORDINATESINFO"]._serialized_start = 299
    _globals["_DJIREMOTEGPSCOORDINATESINFO"]._serialized_end = 391
    _globals["_COORDINATESDATETIME"]._serialized_start = 393
    _globals["_COORDINATESDATETIME"]._serialized_end = 432
    _globals["_DJIREMOTEGPSCOORDINATES"]._serialized_start = 435
    _globals["_DJIREMOTEGPSCOORDINATES"]._serialized_end = 597
    _globals["_DJIREMOTEDERIVATIVE"]._serialized_start = 599
    _globals["_DJIREMOTEDERIVATIVE"]._serialized_end = 653
    _globals["_DJIREMOTEGPSINFO"]._serialized_start = 656
    _globals["_DJIREMOTEGPSINFO"]._serialized_end = 810
    _globals["_DJIUNKFLOAT_1"]._serialized_start = 812
    _globals["_DJIUNKFLOAT_1"]._serialized_end = 844
    _globals["_DJIUNKINT32_1"]._serialized_start = 846
    _globals["_DJIUNKINT32_1"]._serialized_end = 878
    _globals["_DJIUNKUINT32_1"]._serialized_start = 880
    _globals["_DJIUNKUINT32_1"]._serialized_end = 913
    _globals["_DJIUNKREPBYTES_1"]._serialized_start = 915
    _globals["_DJIUNKREPBYTES_1"]._serialized_end = 956
    _globals["_DJICAMERAACCELINFO"]._serialized_start = 958
    _globals["_DJICAMERAACCELINFO"]._serialized_end = 1024
    _globals["_DJICAMERAINFO"]._serialized_start = 1027
    _globals["_DJICAMERAINFO"]._serialized_end = 1359
    _globals["_GPSFRAMEINFO"]._serialized_start = 1361
    _globals["_GPSFRAMEINFO"]._serialized_end = 1410
    _globals["_DJIGPSINFO"]._serialized_start = 1413
    _globals["_DJIGPSINFO"]._serialized_end = 1541
    _globals["_VIDEOSTREAMINFODETAILS"]._serialized_start = 1544
    _globals["_VIDEOSTREAMINFODETAILS"]._serialized_end = 1675
    _globals["_VIDEOSTREAMINFO"]._serialized_start = 1677
    _globals["_VIDEOSTREAMINFO"]._serialized_end = 1750
    _globals["_GENERICMESSAGE"]._serialized_start = 1753
    _globals["_GENERICMESSAGE"]._serialized_end = 1893
# @@protoc_insertion_point(module_scope)



import struct


class MP4Manager:
    mp4_file = None
    metadata_track_index = 3

    metadata = None

    offsets = []
    sizes = []

    def __init__(self, mp4_file):
        self.mp4_file = mp4_file
        self._read_chunk_offsets_and_sizes()
        self._extract_chunks()

    def get_metadata(self):
        return self.metadata

    def save_metadata(self, output_file):
        with open(output_file, "wb") as f:
            f.write(self.metadata)
        return True

    def _read_box(self, fp):
        """
        Read a box header and return its details.

        :param fp: File pointer.
        :return: Tuple containing box size, box type, and start position.
        """
        start_of_box = fp.tell()
        header = fp.read(8)
        if len(header) < 8:
            return None, None, None

        box_size, box_type = struct.unpack(">I4s", header)
        box_type = box_type.decode("utf-8")

        if box_size == 1:  # Extended size case
            box_size = struct.unpack(">Q", fp.read(8))[0]

        return box_size, box_type, start_of_box

    def _read_chunk_offsets_and_sizes(self):
        """
        Read the chunk offsets and sizes for a specific track from the
        'co64' and 'stsz' boxes.

        """

        with open(self.mp4_file, "rb") as f:
            while True:
                box_size, box_type, start_of_box = self._read_box(f)
                if not box_size:
                    break

                print(f"Found box: {box_type}, size: {box_size}, start: {start_of_box}")

                if box_type == "moov":
                    moov_data = f.read(box_size - (f.tell() - start_of_box))
                    self._parse_moov_for_co64_and_stsz(moov_data)
                    break

                f.seek(start_of_box + box_size)

        return True

    def _parse_moov_for_co64_and_stsz(self, moov_data):
        """
        Parse the 'moov' box to extract chunk offsets and sizes for
        the specified track.

        """
        track_counter = 0

        i = 0
        while i < len(moov_data):
            box_size, box_type = struct.unpack(">I4s", moov_data[i : i + 8])
            box_type = box_type.decode("utf-8")

            if box_size == 1:  # Extended size case
                box_size = struct.unpack(">Q", moov_data[i + 8 : i + 16])[0]
                header_size = 16
            else:
                header_size = 8

            print(f"Parsing moov child box: {box_type}, size: {box_size}")

            if box_type == "trak":
                track_counter += 1
                if track_counter == self.metadata_track_index:
                    trak_data = moov_data[i + header_size : i + box_size]
                    self._parse_trak_for_co64_and_stsz(trak_data)
                    break

            i += box_size

        return True

    def _parse_trak_for_co64_and_stsz(self, trak_data):
        """
        Parse the 'trak' box to find the 'co64' and 'stsz' boxes
        for chunk offsets and sizes.

        :param trak_data: Binary data of the 'trak' box.
        """

        i = 0
        while i < len(trak_data):
            box_size, box_type = struct.unpack(">I4s", trak_data[i : i + 8])
            box_type = box_type.decode("utf-8")

            if box_size == 1:  # Extended size case
                box_size = struct.unpack(">Q", trak_data[i + 8 : i + 16])[0]
                header_size = 16
            else:
                header_size = 8

            print(f"Parsing trak child box: {box_type}, size: {box_size}")

            if box_type == "mdia":
                mdia_data = trak_data[i + header_size : i + box_size]
                self._parse_mdia_for_co64_and_stsz(mdia_data)
                break

            i += box_size

        return True

    def _parse_mdia_for_co64_and_stsz(self, mdia_data):
        """
        Parse the 'mdia' box to find the 'co64' and 'stsz' boxes for
        chunk offsets and sizes.

        :param mdia_data: Binary data of the 'mdia' box.
        """

        i = 0
        while i < len(mdia_data):
            box_size, box_type = struct.unpack(">I4s", mdia_data[i : i + 8])
            box_type = box_type.decode("utf-8")

            if box_size == 1:  # Extended size case
                box_size = struct.unpack(">Q", mdia_data[i + 8 : i + 16])[0]
                header_size = 16
            else:
                header_size = 8

            print(f"Parsing mdia child box: {box_type}, size: {box_size}")

            if box_type == "minf":
                minf_data = mdia_data[i + header_size : i + box_size]
                self._parse_minf_for_co64_and_stsz(minf_data)
                break

            i += box_size

        return True

    def _parse_minf_for_co64_and_stsz(self, minf_data):
        """
        Parse the 'minf' box to find the 'co64' and 'stsz' boxes
        for chunk offsets and sizes.

        :param minf_data: Binary data of the 'minf' box.
        """

        i = 0
        while i < len(minf_data):
            box_size, box_type = struct.unpack(">I4s", minf_data[i : i + 8])
            box_type = box_type.decode("utf-8")

            if box_size == 1:  # Extended size case
                box_size = struct.unpack(">Q", minf_data[i + 8 : i + 16])[0]
                header_size = 16
            else:
                header_size = 8

            print(f"Parsing minf child box: {box_type}, size: {box_size}")

            if box_type == "stbl":
                stbl_data = minf_data[i + header_size : i + box_size]
                self._parse_stbl_for_co64_and_stsz(stbl_data)
                break

            i += box_size

        return True

    def _parse_stbl_for_co64_and_stsz(self, stbl_data):
        """
        Parse the 'stbl' box to find the 'co64' and 'stsz' boxes
        for chunk offsets and sizes.

        :param stbl_data: Binary data of the 'stbl' box.
        """

        i = 0
        while i < len(stbl_data):
            box_size, box_type = struct.unpack(">I4s", stbl_data[i : i + 8])
            box_type = box_type.decode("utf-8")

            if box_size == 1:  # Extended size case
                box_size = struct.unpack(">Q", stbl_data[i + 8 : i + 16])[0]
                header_size = 16
            else:
                header_size = 8

            print(f"Parsing stbl child box: {box_type}, size: {box_size}")

            if box_type == "co64":
                co64_data = stbl_data[i + header_size : i + box_size]
                self._parse_co64(co64_data)
            elif box_type == "stsz":
                stsz_data = stbl_data[i + header_size : i + box_size]
                self._parse_stsz(stsz_data)

            i += box_size

        return True

    def _parse_co64(self, data):
        """
        Parse the 'co64' box for chunk offsets.

        :param data: Binary data of the 'co64' box.
        :return: List of chunk offsets.
        """
        offsets = []

        # Ensure there are at least 12 bytes for the header
        if len(data) < 12:
            raise ValueError(
                f"Insufficient data for 'co64' header. Got {len(data)} "
                "bytes, expected at least 12."
            )

        # Extract flags/version (4 byte) e entry_count (4 byte)
        flags_version, entry_count = struct.unpack(">II", data[:8])
        print(f"Found co64 box with flags={flags_version}, and {entry_count} entries.")

        # Calculate required length and validate
        required_length = 8 + entry_count * 8
        if len(data) < required_length:
            raise ValueError(
                f"Incomplete 'co64' data. Expected {required_length} "
                f"bytes, got {len(data)}."
            )

        # Extract chunk offsets (8 bytes each)
        for i in range(entry_count):
            start = 8 + i * 8
            end = start + 8
            offset = struct.unpack(">Q", data[start:end])[0]
            offsets.append(offset)

        self.offsets = offsets
        return True

    def _parse_stsz(self, data):
        """
        Parse the 'stsz' box for chunk offsets.

        :param data: Binary data of the 'stsz' box.
        :return: List of chunk sizes.
        """
        sizes = []

        # Ensure there are at least 12 bytes for the header
        if len(data) < 12:
            raise ValueError(
                f"Insufficient data for 'stsz' header. Got {len(data)} "
                "bytes, expected at least 12."
            )

        # Extract flags/version (4 byte) e entry_count (4 byte)

        flags, version, entry_count = struct.unpack(">III", data[:12])
        print(
            f"Found stsz box with flags={flags}, version={version}, "
            f"and {entry_count} entries."
        )

        # Calculate required length and validate
        required_length = 4 + entry_count * 4
        if len(data) < required_length:
            raise ValueError(
                f"Incomplete 'stsz' data. Expected {required_length} "
                f"bytes, got {len(data)}."
            )

        # Extract chunk sizes (4 bytes each)
        for i in range(entry_count):
            start = 12 + i * 4
            end = start + 4
            size = struct.unpack(">I", data[start:end])[0]
            sizes.append(size)
        self.sizes = sizes
        return True

    def _extract_chunks(self):
        """
        Extract chunks from the 'mdat' box and join them into a single file.
        """
        with open(self.mp4_file, "rb") as f:
            for i, offset in enumerate(self.offsets):
                f.seek(offset)
                chunk_data = f.read(self.sizes[i])
                self._append_metadata(chunk_data)
        return True

    def _append_metadata(self, chunk_data):
        """
        Append metadata to the extracted chunk data.

        :param chunk_data: Binary data of the extracted chunk.
        """
        if self.metadata:
            chunk_data += self.metadata
        else:
            self.metadata = chunk_data

        return True
